#ifndef CLASSIFIER_H
#define CLASSIFIER_H
#include <qmainwindow.h>
#include <qimage.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qspinbox.h>
#include <qrect.h>
#include <qvalidator.h>
class FloatSpinBox: public QSpinBox
{
    Q_OBJECT
    public:
        FloatSpinBox( QWidget* parent, const char* name ) : QSpinBox( parent, name ){ 
            QSpinBox::setValidator ( new QDoubleValidator( this, "doubleValidator") );
        }
        QString mapValueToText( int value )        {
            return QString("%1.%2")
                        .arg( value / 100 )
                        .arg( value % 100 );
        }
        int mapTextToValue( bool * ok )        {
            return (int)( 100 * text().toFloat() );
        }
};

class QuadTreeNode: public QRect
{
public:
    QuadTreeNode* kids[2][2];
    bool isSplit;
    float distribution;
    QImage* image;

    QuadTreeNode( int x, int y, int w, int h, QImage* _image ):
    QRect(x,y,w,h){ 
        image = _image; 
        isSplit = false;
        distribution = -1;
    }

    ~QuadTreeNode(){
        if( isSplit )
            for( int i = 0; i < 2; i ++ )
                for( int j = 0; j < 2; j ++ )
                    delete kids[i][j];
    }
    
    void build( int thresh, float biash, float biass, float biasv ){
        calcDistribution( biash, biass, biasv );
        if( distribution > thresh && !isSplit )
            split();
        if( isSplit ){
            for( int i = 0; i < 2; i ++ )
                for( int j = 0; j < 2; j ++ )
                    kids[i][j]->build( thresh, biash, biass, biasv );
        }
    }
        
    void split(){
        int x = QRect::x();
        int y = QRect::y();
        int w0 = (int)(width()/2.0f);
        int w1 = (int)(width()/2.0f + .5f);
        int h0 = (int)(height()/2.0f);
        int h1 = (int)(height()/2.0f + .5f);
        kids[0][0] = new QuadTreeNode( x,   y   ,w0,h0, image );
        kids[0][1] = new QuadTreeNode( x+w0,y   ,w1,h0, image );
        kids[1][0] = new QuadTreeNode( x+w0,y+h0,w1,h1, image );
        kids[1][1] = new QuadTreeNode( x,   y+h0,w0,h1, image );
        isSplit = true;
    }
    
    void calcDistribution( float biash, float biass, float biasv ){
        int i,j;
        int h,s,v;
        h = s = v = 0;
        long sh, ss, sv;
        sh = ss = sv = 0;
        int ah, as, av;
        ah = as = av = 0;
        float dh, ds, dv;
        dh = ds = dv = 0;
        int n = width() * height();
        if( n < 2 ) { distribution = 0; return; }
        QColor c;
        for( i = 0; i < width(); i ++ )
            for( j = 0; j < height(); j ++ ){
                c.setRgb( image->pixel(i+x(),j+y()) );
                c.hsv( &h,&s,&v );
                sh += h;
                ss += s;
                sv += v;
        }
        ah = sh / n;
        as = ss / n;
        av = sv / n;

        for( i = 0; i < width(); i ++ )
            for( j = 0; j < height(); j ++ ){
                c.setRgb( image->pixel(i+x(),j+y()) );
                c.hsv( &h,&s,&v );
                dh += biash * abs( h - ah );
                ds += biass * abs( s - as );
                dv += biasv * abs( v - av );
            }
        distribution = dh + ds + dv;
        distribution /= n;
    }

    void paint( QPainter* p ){
        QRect r( x(), y(), width()+1 , height()+1);
        p->drawRect( r );
        if( isSplit )
            for( int i = 0; i < 2; i ++ )
                for( int j = 0; j < 2; j ++ )
                    kids[i][j]->paint( p );
    }

    void numChildren( int& num){
        num++;
        if( isSplit )
            for( int i = 0; i < 2; i ++ )
                for( int j = 0; j < 2; j ++ )
                   kids[i][j]->numChildren(num);
    }
        
};
    
class Classifier : public QMainWindow
{
    Q_OBJECT
public:
    Classifier( QWidget* parent, const char* name, QImage& _image ):
        QMainWindow( parent, name ),
        biass(1.0f), biash(1.0f), biasv(1.0f)
    {
        image = _image;
        l = new QLabel( this, "daImage" );
        l->setPixmap( QPixmap( image ) );
        l->setScaledContents( true );
        setCentralWidget( l );
        setMinimumSize( image.width(), image.height() );
        timer = new QTimer( this, "classifier::updateTimer" );
        connect( timer, SIGNAL( timeout() ), this, SLOT(repaint() ));

        QToolBar *t = new QToolBar( this, "classifier::toolbar");
        QToolButton * run = new QToolButton(
                QIconSet( QPixmap("run.bmp" ) ),
                "Select Point",
                QString::null,
                this, SLOT( build() ),
                t,
                "Run" );

        QLabel* tlbl = new QLabel( t, "t display" );
        tlbl->setText("T: ");
        QSpinBox* b = new QSpinBox( t, "ThreshHold" );
        connect( b, SIGNAL( valueChanged( int )) , this, SLOT( setThresh( int )));
        b->setMinValue( 1 );
        b->setMaxValue( 255*3 );
        b->setValue( 10 );
        t->show();
        head = 0;


        QLabel* albl = new QLabel( t, "a display" );
        albl->setText("A: ");
        FloatSpinBox * biash = new FloatSpinBox( t, "BiasH" );
        connect( biash, SIGNAL( valueChanged( int )), this, SLOT( setbiash( int )));
        biash->setMinValue( 0 );
        biash->setMaxValue( 1000 );
        biash->setValue( 100 );
    
        QLabel* blbl = new QLabel( t, "b display" );
        blbl->setText("B: ");
        FloatSpinBox * biass = new FloatSpinBox( t, "Biass" );
        connect( biass, SIGNAL( valueChanged( int )), this, SLOT( setbiass( int )));
        biass->setMinValue( 0 );
        biass->setMaxValue( 1000 );
        biass->setValue( 100 );

        QLabel* clbl = new QLabel( t, "c display" );
        clbl->setText("C: ");
        FloatSpinBox * biasv = new FloatSpinBox( t, "Biasv" );
        connect( biasv, SIGNAL( valueChanged( int )), this, SLOT( setbiasv( int )));
        biasv->setMinValue( 0 );
        biasv->setMaxValue( 1000 );
        biasv->setValue( 100 );

        QLabel* flbl = new QLabel( t, "func display" );
        flbl->setText("if( Ah + Bs + Cv < T ) split()");
        
        numRects = new QLabel( t, "numRects" );
    }
    ~Classifier(){
        delete head;
    }

public slots:
    void build(){
        timer->start(60);
        if( head )
            delete head;
        head = 0;
        repaint();
        head = new QuadTreeNode( 0,0,image.width(),image.height(), &image );
        head->build( thresh, biash, biass, biasv );
        update();
        timer->stop();
        int i = 0;
        head->numChildren(i);
        numRects->setText( QString("Num rects : %1").arg(i));

    }
    void paintEvent( QPaintEvent* e ){
        l->repaint();
        QPainter* p = new QPainter( l );
        p->setPen( QColor( 0,0,0 ) );
        if( head )
            head->paint(p);
    }
    void setThresh( int a ){
        thresh = a;
    }
    void setbiash( int a ) { biash = ((float) a ) / 100.0f; }
    void setbiass( int a ) { biass = ((float) a ) / 100.0f; }
    void setbiasv( int a ) { biasv = ((float) a ) / 100.0f; }
    void hideEvent( QHideEvent* e ){
        close( true );
    }
    
protected:
    QToolBar *t;
    QLabel * l;
    QLabel * numRects;
    QImage image;
    QuadTreeNode* head;
    int thresh;

    QTimer * timer;
    float biash, biass, biasv;
};
#endif
